// Bug fix: When storing events into internal_events, if note event, remember current vol and pan.

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "findkey.c"


#define DPQN 24
// divisions per quarter note

typedef unsigned char UI1;  // 1-byte unsigned integer
typedef unsigned long UI4;  // 4-byte unsigned integer

enum Event_type_enum {EVENT_UNDEFINED, EVENT_END, EVENT_START, EVENT_CONTROL};

typedef enum Event_type_enum Event_type;

struct Event_struct
{
  UI4 time;
  UI1 track;
  UI1 instrument;
  Event_type type;
  UI1 pitch;       // n/a for type EVENT_CONTROL
  UI1 vol;         // n/a for type EVENT_END
  UI1 pan;         // n/a for type EVENT_END
};

typedef struct Event_struct Event;

int compare_events(const Event* a, const Event* b)
{
  if      (a->time < b->time) return -1;
  else if (a->time > b->time) return  1;
  else if (a->type < b->type) return -1;
  else if (a->type > b->type) return  1;
  else if (a->track < b->track) return -1;
  else if (a->track > b->track) return  1;
  else if (a->instrument < b->instrument) return -1;
  else if (a->instrument > b->instrument) return  1;
  else return 0;
}
                               // C  C#   D  D#  E   F F#   G  G#  A  A#  B
unsigned char key_signature[] = { 0, -5,  2, -3, 4, -1, 6,  1, -4, 3, -2, 5,    // major
                                 -3,  4, -1, -6, 1, -4, 3, -2,  5, 0, -5, 2 };  // minor

unsigned char orgdrum_to_mididrum[] = {35, 36, 38, 40, 50, 44, 46, 49, 37, 76,
                                       35, 45, 35, 36, 38, 40, 42, 46, 42, 46,
                                       57, 49, 51, 47, 41, 38, 83, 83         };

#include "volpanconv.h"

int main(int num_args, char *args[])
{
  FILE *infile, *outfile;
  unsigned long infilesize, pos, pos_beat, pos_pitch, pos_dur, pos_vol, pos_pan, num_internal_events, internal_pos;
  unsigned long skipped_internal_events, last_pos, track_len;
  unsigned int beats[12], i, best_key, num_org_events, total_org_events;
  unsigned char track, channel;
  UI1 current_instrument;
  UI4 current_end, latest_end, loop_start, loop_end;
  unsigned char *buffer;
  Event *internal_events;
  unsigned char out_buffer[256];
  unsigned char time_sig_num, time_sig_denom, org_step_value;
  int org_step_divisions;
  UI4 us_per_qn, current_time, last_time, delta_time;
  unsigned char current_org_vol, current_org_pan, current_midi_vel, current_midi_pan;

  if(num_args != 3)
  {
    fprintf(stderr, "Usage: ORGMID in.org out.mid\n");
    return 1;
  }

  infile = fopen(args[1],"rb");
  if(!infile)
  {
    fprintf(stderr, "Unable to open input file.\n");
    return 1;
  }

  fseek(infile, 0, SEEK_END);
  infilesize = ftell(infile);
  fseek(infile, 0, SEEK_SET);

  buffer = (unsigned char *) malloc(infilesize);
  if(!buffer)
  {
    fprintf(stderr, "Not enough memory to load input file.\n");
    return 1;
  }

  if( fread(buffer, 1, infilesize, infile) < infilesize )
  {
    fprintf(stderr, "Error loading input file.\n");
    return 1;
  }

  fclose(infile);

  if( (infilesize < 114) || ( (memcmp(buffer, "Org-02", 6) != 0) && (memcmp(buffer, "Org-03", 6) != 0) ) )
  {
    fprintf(stderr, "Input file isn't an ORG file.\n");
    free(buffer);
    return 1;
  }

  outfile = fopen(args[2],"rb");
  if(outfile)
  {
    fprintf(stderr, "Output file exists.\n");
    fclose(outfile);
    return 1;
  }

  outfile = fopen(args[2],"wb");
  if(!outfile)
  {
    fprintf(stderr, "Unable to open output file.\n");
    return 1;
  }

  // first pass

  // count number of internal events needed
  //   each ORG event with a note becomes 2 internal events (start and end),
  //   other ORG events become 1 internal event.

  // also prepeare beats count array for key finder

  num_internal_events = 0;

  for(i=0; i<12; ++i)
  {
    beats[i] = 0;
  }

  total_org_events = 0;
  for( track=0; track<16; ++track )
  {
    pos = 22 + track*6;
    num_org_events = buffer[pos] + 256*buffer[pos + 1];

    num_internal_events += num_org_events * 2; // For starting purposes, assume all ORG events have a note
                                               //   and will have a start event and an end event.

    pos_pitch = 114 + 8*total_org_events + 4*num_org_events; // beginning of note pitches for current track
    pos_dur   = 114 + 8*total_org_events + 5*num_org_events; // beginning of note durations for current track
    for( i=0; i<num_org_events; ++i )
    {
      if( buffer[pos_pitch] != 255 )  // if ORG event has a note
      {
        if( track < 8 )  // if track is meoldic (only first 8 tracks are melodic, last 8 tracks are precussion)
        {
          beats[ buffer[pos_pitch] % 12 ] += buffer[pos_dur];
        }
      }
      else  // else event only has vol/pan control changes
      {
        --num_internal_events;  // control events only use 1 internal event not 2 as assumed above, so decrease by 1.
      }
      ++pos_pitch; ++pos_dur;
    }
    total_org_events += num_org_events;
  }

  // second pass
  // load events into an array of internal representation events
  // also find latest end

  internal_events = (Event *) malloc(num_internal_events * sizeof(Event) );
  skipped_internal_events = 0;
  internal_pos = 0;
  latest_end = 0;
  current_org_vol = 255;
  current_org_pan = 255;

  total_org_events = 0;
  for( track=0; track<16; ++track )
  {
    pos = 22 + track*6;
    num_org_events = buffer[pos] + 256*buffer[pos + 1];

    pos = 20 + track*6;
    current_instrument = buffer[pos];
    if( track >= 8 ) current_instrument += 100; // percussion instruments

    pos_beat   = 114 + 8*total_org_events;      // beginning of note positions for current track
    pos_pitch  = pos_beat  + 4*num_org_events;  // beginning of note pitches for current track
    pos_dur    = pos_pitch +   num_org_events;  // beginning of note durations for current track
    pos_vol    = pos_dur   +   num_org_events;  // beginning of note volumes for current track
    pos_pan    = pos_vol   +   num_org_events;  // beginning of note pans for current track

    for( i=0; i<num_org_events; ++i )
    {
      internal_events[internal_pos].time =    buffer[pos_beat]
                                          + ( buffer[pos_beat + 1] <<  8)
                                          + ( buffer[pos_beat + 2] << 16)
                                          + ( buffer[pos_beat + 3] << 24);

      internal_events[internal_pos].track = track;
      internal_events[internal_pos].instrument = current_instrument;

      if( buffer[pos_pitch] == 255 )  // if ORG event has no note
      {
        if(    (buffer[pos_vol] == current_org_vol)
            && (buffer[pos_pan] == current_org_pan) ) // if ORG volume and pan in event are same as current
        {
          ++skipped_internal_events;
        }
        else // else ORG volume changed or ORG pan changed
        {
          current_org_vol = buffer[pos_vol];
          current_org_pan = buffer[pos_pan];

          internal_events[internal_pos].type = EVENT_CONTROL;
          internal_events[internal_pos].pitch = 255;
          internal_events[internal_pos].vol = current_org_vol;
          internal_events[internal_pos].pan = current_org_pan;
          ++internal_pos;
        }
      }
      else  // else ORG event has a note
      {
        internal_events[internal_pos].type = EVENT_START;
        internal_events[internal_pos].pitch = buffer[pos_pitch];
        internal_events[internal_pos].vol = buffer[pos_vol];
        internal_events[internal_pos].pan = buffer[pos_pan];

        if( buffer[pos_vol] != 255 ) current_org_vol = buffer[pos_vol];
        if( buffer[pos_pan] != 255 ) current_org_pan = buffer[pos_pan];

        ++internal_pos;

        current_end = internal_events[internal_pos-1].time + buffer[pos_dur];
        if( latest_end < current_end ) latest_end = current_end;

        internal_events[internal_pos].time = current_end;
        internal_events[internal_pos].track = track;
        internal_events[internal_pos].instrument = current_instrument;
        internal_events[internal_pos].type = EVENT_END;
        internal_events[internal_pos].pitch = buffer[pos_pitch];
        internal_events[internal_pos].vol = 255;
        internal_events[internal_pos].pan = 255;
        ++internal_pos;
      }
      pos_beat += 4; ++pos_pitch; ++pos_dur; ++pos_vol; ++pos_pan;
    }
    total_org_events += num_org_events;
  }

  loop_start = buffer[10] + (buffer[11] << 8) + (buffer[12] << 16) + (buffer[13] << 24);
  loop_end   = buffer[14] + (buffer[15] << 8) + (buffer[16] << 16) + (buffer[17] << 24);

  num_internal_events -= skipped_internal_events;

  qsort(internal_events, num_internal_events, sizeof(Event), compare_events);

  best_key = findkey(beats);

  // ==========================

  memcpy(out_buffer, "MThd", 4);

  // length of header is 6
  out_buffer[4] = 0;
  out_buffer[5] = 0;
  out_buffer[6] = 0;
  out_buffer[7] = 6;

  // type 1
  out_buffer[8] = 0;
  out_buffer[9] = 1;

  // 1 conductor track + 16 tracks
  out_buffer[10] = 0;
  out_buffer[11] = 17;

  // divisions per quarter note
  out_buffer[12] = 0;
  out_buffer[13] = DPQN;

  fwrite(out_buffer, 1, 14, outfile);

  memcpy(out_buffer, "MTrk", 4);

  // chunk length place holder
  out_buffer[4] = 0;
  out_buffer[5] = 0;
  out_buffer[6] = 0;
  out_buffer[7] = 0;

  // buffer[8] = beats per measure
  // buffer[9] = steps per beat
  time_sig_num = buffer[8];
//  if( (buffer[9] % 3) == 0 )
//  {
//   time_sig_num *= 3;
//    time_sig_denom = 8;
//    org_step_value = 8 * (buffer[9] / 3);
//  }
//  else
//  {
    time_sig_denom = 4;
    org_step_value = 4 * buffer[9];
//  }
  // org_step_value is the note value of the OrgMaker step:
  //   1=whole note, 2=half note, 4=quarter note, 8=eighth note, etc.


  //time signature
  out_buffer[8]  = 0;
  out_buffer[9]  = 0xFF;
  out_buffer[10] = 0x58;
  out_buffer[11] = 0x04;

  //time signature numerator
  out_buffer[12] = time_sig_num;

  //time signature denomenator
  if      (time_sig_denom ==  1) out_buffer[13] = 0;
  else if (time_sig_denom ==  2) out_buffer[13] = 1;
  else if (time_sig_denom ==  4) out_buffer[13] = 2;
  else if (time_sig_denom ==  8) out_buffer[13] = 3;
  else if (time_sig_denom == 16) out_buffer[13] = 4;
  else if (time_sig_denom == 32) out_buffer[13] = 5;
  else
  {
    fprintf(stderr, "Unexpected time signature denominator.\n");
    return 1;
  }

  // metronome click period in units of 1/24 quarter note
  out_buffer[14] = 24 * 4 * buffer[9] / org_step_value;   // set metronome click period to length of OrgMaker beat

  // always eight (size of 32nd note in units of 1/24 quarter note)
  out_buffer[15] = 8;


  //tempo
  us_per_qn = 1000 * ( buffer[6] + (buffer[7] << 8) ) / 4 * org_step_value;

  out_buffer[16] = 0;
  out_buffer[17] = 0xFF;
  out_buffer[18] = 0x51;
  out_buffer[19] = 0x03;
  out_buffer[20] = (unsigned char) ( (us_per_qn & 0xFF0000) >> 16);
  out_buffer[21] = (unsigned char) ( (us_per_qn &   0xFF00) >>  8);
  out_buffer[22] = (unsigned char) (  us_per_qn &     0xFF       );

  //key signature
  out_buffer[23] = 0;
  out_buffer[24] = 0xFF;
  out_buffer[25] = 0x59;
  out_buffer[26] = 0x02;
  out_buffer[27] = key_signature[best_key];
  out_buffer[28] = (best_key < 12)? 0 : 1;

  fwrite(out_buffer, 1, 29, outfile);

  // loop start marker

  org_step_divisions = 4 * DPQN / org_step_value;
  delta_time = loop_start * org_step_divisions;

  pos = 0;
  if (        delta_time & 0xFE00000 ) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time & 0xFE00000) >> 21) );
  if (pos || (delta_time &  0x1FC000)) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time &  0x1FC000) >> 14) );
  if (pos || (delta_time &    0x3F80)) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time &    0x3F80) >>  7) );
                                       out_buffer[pos++] = (unsigned char) (          delta_time &      0x7F         );
  fwrite(out_buffer, 1, pos, outfile);

  out_buffer[0] = 0xFF;
  out_buffer[1] = 0x06;
  out_buffer[2] = 9;
  memcpy(&(out_buffer[3]), "loopStart", 9);
  fwrite(out_buffer, 1, 12, outfile);


  // loop end marker

  delta_time = (loop_end * org_step_divisions + 1) - delta_time;

  pos = 0;
  if (        delta_time & 0xFE00000 ) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time & 0xFE00000) >> 21) );
  if (pos || (delta_time &  0x1FC000)) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time &  0x1FC000) >> 14) );
  if (pos || (delta_time &    0x3F80)) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time &    0x3F80) >>  7) );
                                       out_buffer[pos++] = (unsigned char) (          delta_time &      0x7F         );
  fwrite(out_buffer, 1, pos, outfile);

  out_buffer[0] = 0xFF;
  out_buffer[1] = 0x06;
  out_buffer[2] = 7;
  memcpy(&(out_buffer[3]), "loopEnd", 7);
  fwrite(out_buffer, 1, 10, outfile);


  //end of track
  out_buffer[0] = 0;
  out_buffer[1] = 0xFF;
  out_buffer[2] = 0x2F;
  out_buffer[3] = 0;

  fwrite(out_buffer, 1, 4, outfile);

  for(track = 0; track < 16; ++track)
  {
    current_time = 0; last_time = 0; delta_time = 0;

    if(track < 8) channel = track;
    else          channel = 9; // "channel 10" (1-16) = 9 (0-15)

    pos = 0;

    memcpy( out_buffer, "MTrk", 4); pos += 4;
    // chunk length place holder
    out_buffer[4] = 0;
    out_buffer[5] = 0;
    out_buffer[6] = 0;
    out_buffer[7] = 0;
    fwrite(out_buffer, 1, 8, outfile);

    if(track < 8)
    {
      // program change - organ
      out_buffer[0] = 0;
      out_buffer[1] = (0xC0 | channel);
      out_buffer[2] = 16;
      fwrite(out_buffer, 1, 3, outfile);
    }

    // vol - max
    out_buffer[0] = 0;
    out_buffer[1] = (0xB0 | channel);
    out_buffer[2] = 7;
    out_buffer[3] = 127;
    fwrite(out_buffer, 1, 4, outfile);

    // default org vol and pan to default midi vel and pan
    current_org_vol = 248;
    current_org_pan = 6;
    current_midi_vel = org_vol_pan_to_midi_vel[current_org_vol][current_org_pan];
    current_midi_pan = org_pan_to_midi_pan[current_org_pan];

    // default midi vel (expression)
    out_buffer[0] = 0;
    out_buffer[1] = (0xB0 | channel);
    out_buffer[2] = 11;
    out_buffer[3] = current_midi_vel;
    fwrite(out_buffer, 1, 4, outfile);

    // default midi pan
    out_buffer[0] = 0;
    out_buffer[1] = (0xB0 | channel);
    out_buffer[2] = 10;
    out_buffer[3] = current_midi_pan;
    fwrite(out_buffer, 1, 4, outfile);

    for(internal_pos = 0; internal_pos < num_internal_events; ++internal_pos)
    {
      if(internal_events[internal_pos].track == track)
      {
        current_time = internal_events[internal_pos].time * org_step_divisions;
        delta_time = current_time - last_time;

        pos = 0;
        if (        delta_time & 0xFE00000 ) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time & 0xFE00000) >> 21) );
        if (pos || (delta_time &  0x1FC000)) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time &  0x1FC000) >> 14) );
        if (pos || (delta_time &    0x3F80)) out_buffer[pos++] = (unsigned char) ( 0x80 | ((delta_time &    0x3F80) >>  7) );
                                             out_buffer[pos++] = (unsigned char) (          delta_time &      0x7F         );
        fwrite(out_buffer, 1, pos, outfile);

        if(internal_events[internal_pos].type == EVENT_START )
        {
          if(internal_events[internal_pos].pan <= 12)
          {
            if( current_org_pan != internal_events[internal_pos].pan)
            {
              current_org_pan = internal_events[internal_pos].pan;
              current_midi_pan = org_pan_to_midi_pan[current_org_pan];
              current_midi_vel = org_vol_pan_to_midi_vel[current_org_vol][current_org_pan];

              // pan
              out_buffer[0] = (0xB0 | channel);
              out_buffer[1] = 10;
              out_buffer[2] = current_midi_pan;
              out_buffer[3] = 0; // delta time 0
              fwrite(out_buffer, 1, 4, outfile);

              // vel (expression)
              out_buffer[0] = (0xB0 | channel);
              out_buffer[1] = 11;
              out_buffer[2] = current_midi_vel;
              out_buffer[3] = 0; // delta time 0
              fwrite(out_buffer, 1, 4, outfile);

            }
          }
          if(    (internal_events[internal_pos].vol != 255)
              && (org_vol_pan_to_midi_vel[internal_events[internal_pos].vol][current_org_pan] != current_midi_vel) )
          {
            current_midi_vel = org_vol_pan_to_midi_vel[internal_events[internal_pos].vol][current_org_pan];
            out_buffer[0] = (0xB0 | channel);
            out_buffer[1] = 11;
            out_buffer[2] = current_midi_vel;
            out_buffer[3] = 0; // delta time 0
            fwrite(out_buffer, 1, 4, outfile);
          }

          out_buffer[0] = (0x90 | channel);

          if(track < 8) out_buffer[1] = 24 + internal_events[internal_pos].pitch;
          else          out_buffer[1] = orgdrum_to_mididrum[ internal_events[internal_pos].instrument - 100];

          out_buffer[2] = 127;
          fwrite(out_buffer, 1, 3, outfile);
        }
        else if (internal_events[internal_pos].type == EVENT_END )
        {
          out_buffer[0] = (0x90 | channel);

          if(track < 8) out_buffer[1] = 24 + internal_events[internal_pos].pitch;
          else          out_buffer[1] = orgdrum_to_mididrum[ internal_events[internal_pos].instrument - 100];

          out_buffer[2] = 0;
          fwrite(out_buffer, 1, 3, outfile);
        }
        else if (internal_events[internal_pos].type == EVENT_CONTROL )
        {
          pos = 0;

          if(internal_events[internal_pos].pan <= 12)
          {
            if( current_org_pan != internal_events[internal_pos].pan)
            {
              current_org_pan = internal_events[internal_pos].pan;
              current_midi_pan = org_pan_to_midi_pan[current_org_pan];
              current_midi_vel = org_vol_pan_to_midi_vel[current_org_vol][current_org_pan];

              // pan
              out_buffer[pos++] = (0xB0 | channel);
              out_buffer[pos++] = 10;
              out_buffer[pos++] = current_midi_pan;
              out_buffer[pos++] = 0; // delta time 0

              // vel (expression)
              out_buffer[pos++] = (0xB0 | channel);
              out_buffer[pos++] = 11;
              out_buffer[pos++] = current_midi_vel;

            }
          }
          if(    (internal_events[internal_pos].vol != 255)
              && (org_vol_pan_to_midi_vel[internal_events[internal_pos].vol][current_org_pan] != current_midi_vel) )
          {
            current_midi_vel = org_vol_pan_to_midi_vel[internal_events[internal_pos].vol][current_org_pan];

            if(pos) out_buffer[pos++] = 0; // delta_time 0
            out_buffer[pos++] = (0xB0 | channel);
            out_buffer[pos++] = 11;
            out_buffer[pos++] = current_midi_vel;
          }

          if(!pos)
          {
            // Empty or unneeded control event.
            out_buffer[pos++] = (0xB0 | channel);
            out_buffer[pos++] = 11;
            out_buffer[pos++] = current_midi_vel;
          }
          fwrite(out_buffer, 1, pos, outfile);
        }
//      else if (internal_events[internal_pos].type == EVENT_UNDEFINED )
//      {
//        fprintf(stderr,"Event of type EVENT_UNDEFINED encountered.\n");
//      }
//      else
//      {
//        fprintf(stderr,"Event of unknown type encountered.\n");
//      }


        last_time = current_time;
      }
    }

    //end of track
    out_buffer[0] = 0;
    out_buffer[1] = 0xFF;
    out_buffer[2] = 0x2F;
    out_buffer[3] = 0;
    fwrite(out_buffer, 1, 4, outfile);
  }

  free(buffer);
  free(internal_events);
  fclose(outfile);

  // RELOAD OUTPUT FILE, ADD TRACK LENGTHS

  // reopen output file, load into buffer

  infile = fopen(args[2],"rb");
  if(!infile)
  {
    fprintf(stderr, "Unable to reopen output file.\n");
    return 1;
  }

  fseek(infile, 0, SEEK_END);
  infilesize = ftell(outfile);
  fseek(infile, 0, SEEK_SET);

  buffer = (unsigned char *) malloc(infilesize);
  if(!buffer)
  {
    fprintf(stderr, "Not enough memory to reload output file.\n");
    return 1;
  }

  if( fread(buffer, 1, infilesize, infile) < infilesize )
  {
    fprintf(stderr, "Error reloading output file.\n");
    return 1;
  }

  fclose(infile);

  // add track lenghts into buffer

  last_pos = 0;
  for( pos = 3; pos < infilesize; ++pos)
  {
    if(    (buffer[pos-3] == 'M')
        && (buffer[pos-2] == 'T')
        && (buffer[pos-1] == 'r')
        && (buffer[pos  ] == 'k') )
    {
      if(last_pos != 0)
      {
        track_len = pos - last_pos - 8;
        buffer[last_pos + 1] = (unsigned char) ((track_len & 0xFF000000) >> 24);
        buffer[last_pos + 2] = (unsigned char) ((track_len &   0xFF0000) >> 16);
        buffer[last_pos + 3] = (unsigned char) ((track_len &     0xFF00) >>  8);
        buffer[last_pos + 4] = (unsigned char) ( track_len &       0xFF       );
      }

      last_pos = pos;
    }
  }

  track_len = infilesize - last_pos - 5;
  buffer[last_pos + 1] = (unsigned char) ((track_len & 0xFF000000) >> 24);
  buffer[last_pos + 2] = (unsigned char) ((track_len &   0xFF0000) >> 16);
  buffer[last_pos + 3] = (unsigned char) ((track_len &     0xFF00) >>  8);
  buffer[last_pos + 4] = (unsigned char) ( track_len &       0xFF       );
  
  // dump buffer back into output file

  outfile = fopen(args[2],"wb");
  if(!outfile)
  {
    fprintf(stderr, "Unable to open output file (second output).\n");
    return 1;
  }

  if ( fwrite(buffer, 1, infilesize, outfile) != infilesize)
  {
    fprintf(stderr, "Unable to dump output file (second output).\n");
    return 1;
  }

  fclose(outfile);
  free(buffer);

  return 0;
}
